/*****
*
* Copyright (C) 2000 Yoann Vandoorselaere <yoann@mandrakesoft.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
*****/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <assert.h>

#include "detect.h"
#include "rules.h"
#include "rules-type.h"


static plugin_detect_t plugin;
static subscribtion_t subscribtion[] = {
        { p_tcp, NULL },
        { p_udp, NULL },
        { p_end, NULL }
};


typedef struct cnx_infos 
{        
        uint8_t port[65536 / 8];
        unsigned int cnx_count;
        unsigned int first_port;
        unsigned int last_port;

        packet_container_t *packet;
        
        char *kind;
        long first_cnxtime;
        long last_cnxtime;
        prelude_timer_t timer;
} cnx_infos_t;


static unsigned int cnx_ttl = 0;
static unsigned int max_cnx_count = 0;
static unsigned int plug_id;
static ip_t *ignore = NULL;



/*
 *
 */
static void do_report_if_needed(cnx_infos_t *cnx, packet_container_t *packet) 
{
        if ( cnx->cnx_count >= max_cnx_count ) {
                struct timeval tv;
                
                gettimeofday(&tv, NULL);
                
                plugin_do_report((plugin_generic_t *) &plugin,
                                 packet, normal, "Scanning attack",
                                 "%s scanning attempt: %u cnx from port "
                                 "%u to %u in %ld seconds.",
                                 cnx->kind, cnx->cnx_count,
                                 cnx->first_port, cnx->last_port,
                                 tv.tv_sec - cnx->first_cnxtime);
        }    
}



/*
 *
 */
static void expire_cnx(void *arg) 
{
        hostdb_t *h = arg;
        cnx_infos_t *cnx = hostdb_get_plugin_data(h, plug_id);
        
        timer_destroy(&cnx->timer);

        do_report_if_needed(cnx, cnx->packet);

        packet_release(cnx->packet);
        hostdb_del(h, plug_id);
        free(cnx->kind);
        free(cnx);
}



/*
 *
 */
static const char *guess_tcp_scan_kind(unsigned char flags) 
{
        if ( flags & TH_ACK )
                switch(flags) {
                        
                case (TH_SYN | TH_RST | TH_ACK | TH_FIN | TH_PUSH | TH_URG):
                        return "Xmas";
                default:
                        return NULL;

                }
        
        
        switch(flags) {
                
        case TH_RST:
                break;
                
        case TH_SYN :
                return "Syn";
                
        case TH_FIN:
                return "Fin";
                
        case 0:
                return "Null";
                                               
        case (TH_FIN | TH_PUSH | TH_URG):
                return "Xmas";
                
        default:
                return "Unknow";

        }

        return NULL;
}



/*
 *
 */
static cnx_infos_t *new_cnx(packet_container_t *p, hostdb_t *h,  const char *kind, uint16_t dport) 
{
        cnx_infos_t *tmp;
        struct timeval tv;
        
        tmp = malloc(sizeof(cnx_infos_t));
        if ( ! tmp ) {
                log(LOG_ERR, "memory exhausted.");
                return NULL;
        }
        
        gettimeofday(&tv, NULL);
        memset(tmp->port, 0, sizeof(tmp->port));

        /*
         * dport / 8 give use the key to access our array.
         * 0x01 << (dport % 8) set the bit corresponding to dport to 1. 
         */
        tmp->port[dport / 8] |= (0x01 << (dport % 8));
        
        tmp->last_cnxtime = tmp->first_cnxtime = tv.tv_sec;
        tmp->last_port = tmp->first_port = dport;
        tmp->cnx_count = 1;
        tmp->kind = strdup(kind);
        tmp->packet = p;

        packet_lock(p);
        
        timer_set_expire(&tmp->timer, cnx_ttl);
        timer_set_data(&tmp->timer, h);
        timer_set_callback(&tmp->timer, expire_cnx);
        
        timer_init(&tmp->timer);

        return tmp;
}



/*
 *
 */
static void modify_cnx(cnx_infos_t *cnx, uint16_t dport) 
{
        int ret;
        struct timeval tv;
        
        gettimeofday(&tv, NULL);

        cnx->last_cnxtime = tv.tv_sec;

        if ( dport < cnx->first_port )
                cnx->first_port = dport;
        
        else if ( dport > cnx->last_port )
                cnx->last_port = dport;

        /*
         * dport / 8 give use the key to access our array.
         * 0x01 << (dport % 8) set the bit corresponding to dport to 1. 
         */
        ret = cnx->port[dport / 8] & (0x01 << (dport % 8));
        if ( ret == 0 ) {
                cnx->port[dport / 8] |= (0x01 << (dport % 8));
                cnx->cnx_count++;
                timer_reset(&cnx->timer);
        }
}



/*
 *
 */
static void update_hdb_entry(hostdb_t *h, packet_container_t *p, uint16_t dport, const char *kind) 
{
        cnx_infos_t *tmp;

        /*
         * If this plugin didn't registered any data in the host
         * database at this time, create a new cnxInfo_t structure.
         */
        tmp = (cnx_infos_t *) hostdb_get_plugin_data(h, plug_id);
        if ( ! tmp ) {
                tmp = new_cnx(p, h, kind, dport);
                if ( ! tmp )
                        return;
                
                hostdb_set_plugin_data(h, plug_id, tmp);
        } else 
                modify_cnx(tmp, dport); 
}



/*
 *
 */
static void create_hdb_entry(packet_container_t *p, const iphdr_t *ip, uint16_t dport, const char *kind)
{
        hostdb_t *hdb;
        cnx_infos_t *tmp;
        
        hdb = hostdb_new(p, ip);
        if (! hdb)
                return;

        tmp = new_cnx(p, hdb, kind, dport);
        if ( ! tmp )
                return;
        
        hostdb_set_plugin_data(hdb, plug_id, tmp);
}



/*
 *
 */
static void generic_packet(packet_container_t *p, const iphdr_t *ip, uint16_t dport, const char *kind)
{
        hostdb_t *hdb;
        
        if ( dport <= 0 || dport > 65535 )
                return;

        hdb = hostdb_search(ip);
        if ( ! hdb ) 
                create_hdb_entry(p, ip, dport, kind);
        else 
                update_hdb_entry(hdb, p, dport, kind);
}




static void tcp_packet(packet_container_t *p, const iphdr_t *ip, const tcphdr_t *tcp) 
{
        const char *kind;

        kind = guess_tcp_scan_kind(tcp->th_flags);
        if ( ! kind )
                return;

        generic_packet(p, ip, ntohs(tcp->th_dport), kind);
}




static void udp_packet(packet_container_t *p, const iphdr_t *ip, const udphdr_t *udp) 
{        
        generic_packet(p, ip, ntohs(udp->uh_dport), "Udp");
}




/*
 *
 */
static void scandetect_run(packet_container_t *packet) 
{
        const iphdr_t *ip;
        int ip_depth = packet->network_layer_depth;

        if ( ip_depth < 0 || packet->packet[ip_depth].proto != p_ip )
                return;

        ip = packet->packet[ip_depth].p.ip;
        if ( ignore && (ip->ip_src.s_addr & ignore->netmask) == ignore->ip ) {
                struct in_addr addr;
                addr.s_addr = ignore->ip;
                printf("This host (%s) is ignored.\n", inet_ntoa(addr));
                return;
        }
        
        switch (packet->packet[packet->depth].proto) {
                
        case p_tcp:
                tcp_packet(packet, ip, packet->packet[packet->depth].p.tcp);
                break;

        case p_udp:
                udp_packet(packet, ip, packet->packet[packet->depth].p.udp_hdr);
                break;

        default:
                log(LOG_ERR, "Unknow protocol %d.\n", packet->packet[packet->depth].proto);
                return;
        }
}



static void print_help(const char *args) 
{
        fprintf(stderr, "Usage for %s :\n", plugin.name);
        fprintf(stderr, "\t -i --ignore-host Host to ignore.\n");
        fprintf(stderr, "\t -c --cnx-count (default=%d) "
                "Number of cnx to track before issuing report.\n", max_cnx_count);
        fprintf(stderr, "\t -t --cnx-ttl (default=%d) "
                "The time under witch cnx-count must occur.\n\n",
                cnx_ttl);
}



static void set_ignore_host(const char *args) 
{        
        ignore = parse_ip(optarg);
}



static void set_cnx_count(const char *args) 
{
        max_cnx_count = atoi(args);
}



static void set_ttl(const char *args) 
{
        cnx_ttl = atoi(args);
}



int plugin_init(unsigned int id)
{
        plugin_option_t opts[] = {
                { "ignore-host", required_argument, NULL, 'i', set_ignore_host },
                { "cnx-count", required_argument, NULL, 'c', set_cnx_count     },
                { "cnx-ttl", required_argument, NULL, 't', set_ttl             },
                { "help", no_argument, NULL, 'h', print_help                   },
                { 0, 0, 0, 0 },
        };
        
        plug_id = id;
        
        plugin_set_name(&plugin, "ScanDetect");
        plugin_set_author(&plugin, "Yoann Vandoorselaere");
        plugin_set_contact(&plugin, "yoann@mandrakesoft.com");
        plugin_set_desc(&plugin, "Detect almost all kind of scanning.");
        plugin_set_running_func(&plugin, scandetect_run);
        plugin_set_subscribtion(&plugin, subscribtion);

        plugin_config_get((plugin_generic_t *) &plugin, opts, PRELUDE_CONF);
        
        return plugin_register((plugin_generic_t *)&plugin);
}












