/*****
*
* Copyright (C) 2001 Jeremie Brebec <flagg@ifrance.com>
* All Rights Reserved
*
* This file is part of the Prelude program.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by 
* the Free Software Foundation; either version 2, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; see the file COPYING.  If not, write to
* the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Written by Jeremie Brebec <flagg@ifrance.com>
*
*****/

#include <stdio.h>
#include "detect.h"

#include "rules.h"
#include "rules_operations.h"
#include "rules_parsing.h"
#include "rules_default.h"
#include "variable.h"

#define RULE_RQUEUE 10

static int parse_signature_file(const char *file, int level);


static int n_rules, n_ignored;
static plugin_detect_t plugin;
static subscribtion_t subscribtion[] = {
        { p_end, NULL },
};




/*
 * This function is called by the signature engine
 * when a signature match a packet.
 */
static void signature_matched_cb(packet_container_t *packet, data_t *data)
{
        int msg_len;
	char *msg, *ref;
        static rqueue_t report;
        rqueue_t *rqueue = signature_engine_get_data_by_id(data, RULE_RQUEUE);
	data_msg_t *data_msg = signature_engine_get_data_by_id(data, RULE_MESSAGE);
	data_msg_t *data_ref = signature_engine_get_data_by_id(data, RULE_REFERENCE);
        
	if ( ! data_msg ) {
		msg = "Signature matched";
		msg_len = sizeof(msg);
	} else { 
		msg = data_msg->msg;
		msg_len = data_msg->len;
	}

	if ( ! data_ref )
		ref = "";
	else
		ref = data_ref->msg;

	if ( ! rqueue )
		rqueue = &report;
	
	rqueue_report(rqueue, (plugin_generic_t *) &plugin, packet,
                      guess, msg, msg_len + 1, "%s\n\n%s\n", msg, ref);
}



static rules_node_t *get_protocol_node(const char *file, int line) 
{
        const char *param;
        rules_node_t *root = NULL;
        
        param = strtok(NULL, " ");
        if ( ! param ) {
                log(LOG_INFO, "%s (%d) Missing protocole for alert rule\n", file, line);
                return NULL;
        }

        if ( strcasecmp(param, "ip") == 0 )
                root = signature_engine_get_ip_root();
        
        else if ( strcasecmp(param, "tcp") == 0 )
                root = signature_engine_get_tcp_root();

        else if ( strcasecmp(param, "udp") == 0 )
                root = signature_engine_get_udp_root();
        
        else if ( strcasecmp(param, "icmp") == 0 )
                root = signature_engine_get_icmp_root();

        else 
                log(LOG_INFO, "%s (%d) Unknown or unsupported protocole %s\n", file, line, param);

        return root;
}



/*
 * yacc/lex parser use this global variable to
 * communicate with us.
 */
rules_t *rule_parsed;


static int parse_signature(const char *file, int line, int level, run_f_t match) 
{
        int ret;
        const char *value;        
        rules_node_t *root;
        
        root = get_protocol_node(file, line);
        if ( ! root ) {
                log(LOG_ERR, "couldn't get protocol node.\n");
                return -1;
        }
        
        value = strtok(NULL, "");
        if ( ! value ) {
                log(LOG_INFO, "%s (%d) Missing test\n", file, line);
                return -1;
        }
        
        set_parsing_buffer(value);
        
	ret = yyparse();
        if ( ret != 0 ) {
                signature_parser_set_error("Syntax Error [fix grammar.y to report correctly this error]");
                log(LOG_INFO, "%s (%d) Parse error: %s\n", file, line, signature_parser_get_error_buffer());
                n_ignored++;
                return -1;
        }

        ret = signature_parser_post_processing(rule_parsed);
        if ( ret < 0 ) {
                log(LOG_INFO, "%s (%d) Parse error: %s\n", file, line, signature_parser_get_error_buffer());
                return -1;
        }
        
        
        if (match) {
                rqueue_t *rqueue;

                rqueue = malloc(sizeof(rqueue_t));
                if ( ! rqueue ) {
                        log(LOG_ERR, "memory exhausted.\n");
                        return -1;
                }
                        
                rqueue->report = NULL;
                rule_parsed = rule_and(rule_parsed, make_data_rule(RULE_RQUEUE, rqueue));
        }

        ret = signature_engine_add_rules(root, rule_parsed, match);
        if ( ret < 0 ) {
                delete_rules(rule_parsed);
                return -1;
        }
        
        n_rules++;
        
        return 0;
}




static int parse_include(const char *file, int line, int level) 
{
        const char *param;
        
        /*
         * todo, test of cycle, find the file, etc
         */

        param = strtok(NULL, "");
        if ( ! param ) {
                log(LOG_INFO, "%s (%d) Missing include filename.\n", file, line);
                return -1;
        }
        
        return parse_signature_file(param, level + 1);
}




static int parse_var(const char *file, int line, int level) 
{
        const char *param, *value;
        
        param = strtok(NULL, " ");
        if ( ! param ) {
                log(LOG_INFO, "%s (%d) Missing variable name.\n", file, line);
                return -1;
        }

        value = strtok(NULL, "");
        if ( ! value ) {
                log(LOG_INFO, "%s (%d) Missing value of variable \"%s\".\n", file, line, param);
                return -1;
        }
        
        variable_set(strdup(param), strdup(value));

        return 0;
}



static int parse_preprocessor(const char *file, int line, int level) 
{
        /*
         * preprocessor configuration, just ignore the line
         */
        return -1;
}



static int parse_output(const char *file, int line, int level)
{
        /*
         * output configuration, just ignore the line
         */
        return -1;
}



static int parse_pass(const char *file, int line, int level)
{
        return parse_signature(file, line, level, NULL);
}



static int parse_log(const char *file, int line, int level) 
{
        /* 
         * not implemented
         * i dont know if it can be implemented :-)
         */
        n_ignored++;

        log(LOG_INFO, "%s (%d) Log rule not implemented yet\n", file, line);

        return -1;
}



static int parse_dynamic(const char *file, int line, int level) 
{
        /*
         * not implemented and this is a special log rule, so...
         */
        n_ignored++;

        log(LOG_INFO, "%s (%d) Dynamic rule not implemented yet.\n", file, line);
        
        return -1;
}



static int parse_alert(const char *file, int line, int level) 
{
	return parse_signature(file, line, level, signature_matched_cb);
}



static void parse_line(const char *file, char *line, int linenum, int level)
{
        int i;
        char *cmd;
        static struct {
                const char *rule;
                int (*handler)(const char *file, int line, int level);
        } rtbl[] = {
                { "include", parse_include           },
                { "var", parse_var                   },
                { "preprocessor", parse_preprocessor },
                { "output", parse_output             },
                { "pass", parse_pass                 },
                { "log", parse_log                   },
                { "dynamic", parse_dynamic           },
                { "alert", parse_alert               },
                { NULL, NULL                         },
        };
        
        /*
         * find the beginning of the line
         */
        while( (*line == ' ' || *line == '\t') && *line != '\0' )
                line++;

        /*
         * ignore comment or empty line
         */
        if ( *line == '\0' || *line == '#' || *line == ';' || *line == 0x0a )
                return;
        
        /*
         * delete the last space of the line
         */
        i = strlen(line) - 1;
        while ( i > 0 && (line[i] == ' ' || line[i] == 0x0a)) 
                line[i--] = '\0';

        /*
         * split the line
         */
        cmd = strtok(line, " ");
        if ( ! cmd )
                return;
        
        for ( i = 0; rtbl[i].rule != NULL; i++ ) {
                if ( strcasecmp(rtbl[i].rule, cmd) == 0 ) {
                        rtbl[i].handler(file, linenum, level);
                        return;
                }
        }

        log(LOG_INFO, "%s (%d) Unknow command %s.\n", file, linenum, cmd);
}



static int parse_signature_file(const char *file, int level)
{
	FILE *fd;
        int line = 0;
        char buf[1024];
        
        fd = fopen(file, "r");
        if ( ! fd ) {		
		log(LOG_ERR, "error opening '%s'.\n", file);
		return -1;
	}
        
	while ( fgets(buf, sizeof(buf), fd) ) {

                line++;        
                parse_line(file, buf, line, level);
	}

	fclose(fd);

        return 0;
}
	


static int load_signature_file(const char *filename)
{
        int ret;

	n_rules = n_ignored = 0;
	ret = parse_signature_file(filename, 0);

        
        log(LOG_INFO, "\t\tSignature engine added %d and ignored %d signature.\n", n_rules, n_ignored);

        return ret;
}



static void rules_run(packet_container_t *packet)
{
	/*
         * nothing
         */
}  



static int help_set = 0;



static void print_help(const char *optarg) 
{
        help_set = 1;
	fprintf(stderr, "Usage for %s :\n", plugin_name(&plugin));
	fprintf(stderr, "\t -r --ruleset Name of rules file.\n\n" );
}



static char *ruleset = NULL;



static void set_ruleset(const char *optarg) 
{
        ruleset = strdup(optarg);
}



int plugin_init(unsigned int id)
{
        int ret;
        plugin_option_t opts[] = {
                { "ruleset", required_argument, NULL, 'r', set_ruleset },
                { "help", no_argument, NULL, 'h', print_help           },
                { 0, 0, 0, 0 },
	};
        
	plugin_set_name(&plugin, "SnortRules");
        plugin_set_author(&plugin, "The Prelude Team");
        plugin_set_contact(&plugin, "prelude-devel@lists.sourceforge.net");
        plugin_set_desc(&plugin, "Snort signature parser.");
        plugin_set_running_func(&plugin, rules_run);
        plugin_set_subscribtion(&plugin, subscribtion);

        plugin_config_get((plugin_generic_t *) &plugin, opts, PRELUDE_CONF);

        if ( ! help_set ) {

                if ( ! ruleset ) 
                        ruleset = strdup(CONFIG_DIR "/prelude.rules");
                        
                ret = load_signature_file(ruleset);
                free(ruleset);
                
                if ( ret == 0 )
                        return plugin_register((plugin_generic_t *) &plugin);
        }
        
        return -1;
}      
