/*
 *  Copyright (C) 2000 Yoann Vandoorselaere.
 *
 *  This program is free software; you can redistribute it and/or modify 
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Authors: Yoann Vandoorselaere <yoann@mandrakesoft.com>
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>

#include "socket-op.h"
#include "packet.h"
#include "alert.h"
#include "alert-prv.h"
#include "write-func.h"
#include "common.h"

#ifdef HAVE_SSL
  #include "ssl.h"
#endif


#define do_write(fd, ptr, len) do {                       \
        if ( socket_write(fd, ptr, len, ssl_write) <= 0 ) \
                goto err;                                 \
} while (0)



#ifdef HAVE_SSL
/*
 * OpenSSL do not support writev(),
 * so we end up doing many write...
 */
static int write_raw_packet(int fd, packet_t *p) 
{
        int i;

        for ( i = 0; ; i++ ) {
                do_write(fd, &p[i].len, sizeof(p[i].len));
                do_write(fd, &p[i].proto, sizeof(p[i].proto));
                
                if ( p[i].proto == p_end )
                        break;
		
                do_write(fd, p[i].p.ip, p[i].len);
        }
        
        return 0;

 err:
        log(LOG_ERR, "error writing packet.\n");
        return -1;
}



/**
 * write_raw_report:
 * @fd: A file descriptor to write the report to.
 * @alert: A Prelude generated report.
 *
 * This function write a report on the file descriptor
 * specified by fd. This function is only compiled in
 * when SSL support was enabled, cause SSL doesn't support
 * writev().
 * As a result, this function is very slow cause of the
 * number of write() call it make.
 *
 * use writev_raw_report() which have a lower overhead
 * when possible.
 *
 * Returns: On success the number of bytes written, -1 otherwise.
 */
int write_raw_report(int fd, alert_t *alert) 
{
        int ret;
        plugin_generic_t *p = alert_plugin(alert);
        
        do_write(fd, &plugin_name_len(p), sizeof(plugin_name_len(p)));        
        do_write(fd, &plugin_author_len(p), sizeof(plugin_author_len(p)));
        do_write(fd, &plugin_contact_len(p), sizeof(plugin_contact_len(p)));
        do_write(fd, &plugin_desc_len(p), sizeof(plugin_desc_len(p)));
        do_write(fd, &alert_quickmsg_len(alert), sizeof(alert_quickmsg_len(alert)));
        do_write(fd, &alert_kind(alert), sizeof(alert_kind(alert)));
        do_write(fd, &alert_count(alert), sizeof(alert_count(alert)));
        do_write(fd, &alert_time_start(alert), sizeof(alert_time_start(alert)));
        do_write(fd, &alert_time_end(alert), sizeof(alert_time_end(alert)));
        do_write(fd, &alert_depth(alert), sizeof(alert_depth(alert)));
        do_write(fd, &alert_message_len(alert), sizeof(alert_message_len(alert)));
        
        /*
         * Variable sized string...
         */
        do_write(fd, alert_message(alert), alert_message_len(alert));
        do_write(fd, plugin_name(p), plugin_name_len(p));
        do_write(fd, plugin_author(p), plugin_author_len(p));
        do_write(fd, plugin_contact(p), plugin_contact_len(p));        
        do_write(fd, plugin_desc(p), plugin_desc_len(p));
        do_write(fd, alert_quickmsg(alert), alert_quickmsg_len(alert));

        ret = write_raw_packet(fd, alert_packet(alert)->packet);
        if ( ret < 0 )
                goto err;

        return 0;

  err:
        log(LOG_ERR, "error writing report.\n");
        return -1;
}
#endif



static void writev_raw_packet(packet_t *p, struct iovec *iov, size_t *iov_count) 
{
        int i;
        
        for ( i = 0; ; i++ ) {
            
                iov[*iov_count].iov_len = sizeof(p[i].len);
                iov[*iov_count].iov_base = &p[i].len;
                *iov_count += 1;
                
                iov[*iov_count].iov_len = sizeof(p[i].proto);
                iov[*iov_count].iov_base = &p[i].proto;
                *iov_count += 1;
                
                if ( p[i].proto == p_end )
                        break;

                iov[*iov_count].iov_len = p[i].len;
                iov[*iov_count].iov_base = p[i].p.ip;
                *iov_count += 1;
        }
}




/**
 * writev_raw_report:
 * @fd: A file descriptor to write the report to.
 * @alert: A Prelude generated report.
 *
 * This function write a report on the file descriptor
 * specified by fd. 
 *
 * Returns: On success the number of bytes written, -1 otherwise.
 */
int writev_raw_report(int fd, alert_t *alert) 
{
        int ret;
        size_t iov_count = 17;
        plugin_generic_t *p = alert_plugin(alert);

        struct iovec iov[17 + (MAX_PKTDEPTH * 3)] = {
                { &plugin_name_len(p), sizeof(plugin_name_len(p))               },
                { &plugin_author_len(p), sizeof(plugin_author_len(p))           },
                { &plugin_contact_len(p), sizeof(plugin_contact_len(p))         },
                { &plugin_desc_len(p), sizeof(plugin_desc_len(p))               },
                { &alert_quickmsg_len(alert), sizeof(alert_quickmsg_len(alert)) },
                { &alert_kind(alert), sizeof(alert_kind(alert))                 },
                { &alert_count(alert), sizeof(alert_count(alert))               },
                { &alert_time_start(alert), sizeof(alert_time_start(alert))     },
                { &alert_time_end(alert), sizeof(alert_time_end(alert))         },
                { &alert_depth(alert), sizeof(alert_depth(alert))               },
                { &alert_message_len(alert), sizeof(alert_message_len(alert))   },
                { alert_message(alert), alert_message_len(alert)                },
                { plugin_name(p), plugin_name_len(p)                            },
                { plugin_author(p), plugin_author_len(p)                        },
                { plugin_contact(p), plugin_contact_len(p)                      },
                { plugin_desc(p), plugin_desc_len(p)                            },
                { alert_quickmsg(alert), alert_quickmsg_len(alert)              },
        };
        
        writev_raw_packet(alert_packet(alert)->packet, iov, &iov_count);
        
        while (1) {
                ret = writev(fd, iov, iov_count);
                if ( ret < 0 ) {
                        if ( errno == EINTR )
                                continue;

                        log(LOG_ERR, "writev returned an error.\n");
                        return -1;
                }

                break;
        }

        return ret;
}
                
                
                
                  
                












